require "ev/oldandnewlocation"
require "ev/db"
require "ev/rwd"

RWDFile	= "rwdtsv.rwd"
Testset	= "org/adres.tsv"

dash, alpha, numeric	= ARGV.to_par

tablefile	= (alpha.collect{|s| oldlocation(s)}.shift or Testset)
port		= numeric.shift
remote		= dash.include?("-r")
tsvs		= dash.include?("-db")

if dash.include?("--testset")
  puts File.new(Testset).readlines
  exit
end

def s_to_a(input)
  output	= input.split("\t", -1)

  @table.key.times do |n|
    output[n]	= ""	if output[n].nil?
  end

  output
end

class RWDTSV < RWDialog
  def initialize(rwd, table, tsvs)
    super(rwd)

    @table	= table
    @tsvs	= tsvs

    @rwd_ignore_vars		<< "@table"
    @rwd_ignore_vars		<< "@tsvs"
    @rwd_call_after_back	<< "main"
  end

  def self.file(file, *args)
    new(File.new(file).readlines, *args)
  end

  def main
    @rwd_window	= "main"

    if @tsvs
      files	= []

      Dir.new(oldlocation).each do |file|
        if file =~ /\.tsv$/
          files << File.basename(file.gsub(/\.tsv/, ""))
        end
      end

      files.sort!

      @tabellen	= files.rwd_method("tabel")
    else
      search
    end
  end

  def tabel(tablefile)
    @table	= TSVFile.new(oldlocation("#{tablefile}.tsv"))

    search
  end

  def search
    @rwd_window	= "search"

    @searchform	= @table.rwd_form("search_")
  end

  def lijst
    @rwd_window	= "lijst"

    @eentje	= ""
    @nieuw	= false

    col	= []
    val	= []

    @table.headers.each_index do |n|
      h	= @table.headers[n].downcase
      v	= instance_eval "@search_#{h}"

      if not v.nil?
        col.push n
        val.push v
      end
    end

    table	= @table.find(@find).subset(col, val, nil, (@exact == "on"), RWDEmptyline)

    @lijst	= table.rwd_table("eentje", "\t")
  end

  def nieuw
    @rwd_window	= "nieuw"

    @eentje	= ""
    @nieuw	= true

    @table.headers.each do |k, v|
      if instance_eval("@search_#{k.downcase}").nil?
        instance_eval "@nieuw_#{k.downcase} = ''"
      else
        instance_eval "@nieuw_#{k.downcase} = @search_#{k.downcase}"
      end
    end

    @nieuwform	= @table.rwd_form("nieuw_", s_to_a(@eentje), true)
  end

  def eentje
    @rwd_window	= "eentje"

    if (@eentje.nil? or @eentje.empty?) and not @nieuw
      error("You have to select a row.")
    else
      row	= @table[s_to_a(@eentje)]
      @table.headers.each_index do |n|
        instance_eval("@eentje_#{@table.headers[n].downcase} = row[#{n}]")
      end
    end

    @eentjeform	= @table.rwd_form("eentje_", s_to_a(@eentje), true)
  end

  def create
    @rwd_window	= "create"

    key	= []
    val	= []

    @table.headers.each_index do |n|
      h	= @table.headers[n].downcase
      v	= instance_eval "@nieuw_#{h}"

      v = ""	if v == RWDEmptyline

      key.push v	if n < (@table.key)
      val.push v
    end
    key	= val	if key.empty?

    key.collect!{|s| s.gsub(/\r/, "")}
    val.collect!{|s| s.gsub(/\r/, "")}

    if key.include?("")
      error "The key can't contain empty fields."
    else
      if @table.include?(key)
        error("Record already exists.")
      else
        @table.add(key, val)
        @table.save
      end
    end

    @tekst	= key.join(" ")
  end

  def update
    @rwd_window	= "update"

    @table.delete(s_to_a(@eentje))

    key	= []
    val	= []

    @table.headers.each_index do |n|
      h	= @table.headers[n].downcase
      v	= instance_eval "@eentje_#{h}"

      v = ""	if v == RWDEmptyline

      key.push v	if n < (@table.key)
      val.push v
    end
    key	= val	if key.empty?

    key.collect!{|s| s.gsub(/\r/, "")}
    val.collect!{|s| s.gsub(/\r/, "")}

    if key.include?("")
      error "The key can't contain empty fields."
    else
      @table.add(key, val)
      @table.save
    end

    @tekst	= s_to_a(@eentje).join(" ")
  end

  def delete
    @rwd_window	= "delete"

    @table.delete(s_to_a(@eentje))
    @table.save

    @tekst	= s_to_a(@eentje).join(" ")
  end
end

RWDTSV.file(RWDFile, TSVFile.new(tablefile), tsvs).serve(port, remote ? ".rwduids" : nil)
